# pylint: disable=C0103,W0703,C0415
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc
import copy


import octobot_commons.enums as commons_enums
import octobot_commons.configuration.user_inputs as user_inputs
import octobot_commons.configuration as configuration


class AbstractTentacle:
    """
    The parent class of any OctoBot tentacle
    """

    __metaclass__ = abc.ABCMeta

    ALLOW_SUPER_CLASS_CONFIG = (
        False  # when True, the given tentacle can read its parent class configuration
    )
    USER_INPUT_TENTACLE_TYPE = (
        commons_enums.UserInputTentacleTypes.UNDEFINED
    )  # tentacle type, saved in user inputs
    HISTORIZE_USER_INPUT_CONFIG = (
        False  # when True, user input values can be saved and read from the run data db
    )
    CLASS_UI = user_inputs.UserInputFactory(
        commons_enums.UserInputTentacleTypes.UNDEFINED
    )  # class-level user input factory. Used when initializing user inputs with classmethods

    def __init__(self):
        self.logger = None
        self.UI: user_inputs.UserInputFactory = user_inputs.UserInputFactory(
            self.USER_INPUT_TENTACLE_TYPE
        )
        self.UI.set_tentacle_class(self.__class__).set_tentacle_config_proxy(
            self.get_local_config
        )

    @classmethod
    def get_name(cls) -> str:
        """
        Tentacle name based on class name
        :return: the tentacle name
        """
        return cls.__name__

    @classmethod
    def get_all_subclasses(cls) -> list:
        """
        Return all subclasses of this tentacle
        :return: the subclasses
        """
        subclasses_list = cls.__subclasses__()
        if cls.__subclasses__():
            for subclass in copy.deepcopy(subclasses_list):
                subclasses_list += subclass.get_all_subclasses()
        return subclasses_list

    @classmethod
    def is_configurable(cls):
        """
        Override if the tentacle is allowed to be configured
        """
        return True

    @classmethod
    def get_user_commands(cls) -> dict:
        """
        Return the dict of user commands for this tentacle
        :return: the commands dict
        """
        return {}

    def get_local_config(self):
        """
        Implementation required if cls.HISTORIZE_USER_INPUT_CONFIG is True
        :return: the config of the tentacle
        """
        raise NotImplementedError

    @classmethod
    def create_local_instance(cls, config, tentacles_setup_config, tentacle_config):
        """
        Implementation required if cls.HISTORIZE_USER_INPUT_CONFIG is True
        :param config: the global configuration to give to the tentacle
        :param tentacles_setup_config: the global tentacles setup configuration to give to the tentacle
        :param tentacle_config: the tentacle configuration to give to the tentacle
        :return: a local, aimed to be short-lived, tentacle instance
        """
        raise NotImplementedError

    def init_user_inputs(self, inputs: dict) -> None:
        """
        instance method API for user inputs. Used by load_and_save_user_inputs
        Override if this tentacle has user inputs that should be initialized on a specific instance
        Called right before starting the tentacle, should define all the tentacle's user inputs unless
        those are defined somewhere else.
        """

    @classmethod
    def init_user_inputs_from_class(cls, inputs: dict) -> None:
        """
        classmethod API for user inputs. Used by init_user_inputs_from_class
        Override if this tentacle has user inputs that can be initialized on a class level
        Called by load_user_inputs_from_class, should define all the tentacle user inputs.
        """

    async def load_and_save_user_inputs(self, bot_id: str) -> dict:
        """
        instance method API for user inputs
        Initialize and save the tentacle user inputs in run data
        :return: the filled user input configuration
        """
        return await configuration.load_and_save_user_inputs(self, bot_id)

    @classmethod
    def load_user_inputs_from_class(
        cls, tentacles_setup_config, tentacle_config
    ) -> dict:
        """
        classmethod API for user inputs
        Initialize the tentacle user inputs
        Called by get_raw_config_and_user_inputs
        """
        return configuration.load_user_inputs_from_class(
            cls, tentacles_setup_config, tentacle_config
        )

    @classmethod
    async def get_raw_config_and_user_inputs(
        cls, config, tentacles_setup_config, bot_id
    ):
        """
        :return: the tentacle configuration and its list of user inputs
        """
        if not cls.HISTORIZE_USER_INPUT_CONFIG:
            return configuration.get_raw_config_and_user_inputs_from_class(
                cls, tentacles_setup_config
            )
        return await configuration.get_raw_config_and_user_inputs(
            cls, config, tentacles_setup_config, bot_id
        )

    @classmethod
    def get_tentacle_config_traded_symbols(
        cls, config: dict, reference_market: str
    ) -> list:
        """
        :return: the traded symbols of the tentacle according to its tentacle configuration
        """
        raise NotImplementedError(
            "get_tentacle_config_traded_symbols is not implemented"
        )



# pylint: disable=C0116
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums


class Signal:
    def __init__(self, topic: str, content: dict, **_):
        self.topic: str = topic
        self.content: dict = content

    def to_dict(self) -> dict:
        return {
            octobot_commons.enums.SignalsAttrs.TOPIC.value: self.topic,
            octobot_commons.enums.SignalsAttrs.CONTENT.value: self.content,
        }

    def __str__(self):
        return f"{self.to_dict()}"



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as constants


def add_historical_tentacle_config(
    master_config: dict, config_start_time: float, historical_config: dict
):
    """
    Adds the given historical_config to the master_config historical configurations
    """
    if constants.CONFIG_HISTORICAL_CONFIGURATION not in master_config:
        master_config[constants.CONFIG_HISTORICAL_CONFIGURATION] = []
    # use list to ensure it still can be serialized
    master_config[constants.CONFIG_HISTORICAL_CONFIGURATION].append(
        [config_start_time, historical_config]
    )
    # always keep the most recent first to be able to find the most up to date first when iterating
    master_config[constants.CONFIG_HISTORICAL_CONFIGURATION].sort(
        key=lambda x: x[0], reverse=True
    )


def get_historical_tentacle_config(master_config: dict, current_time: float) -> dict:
    """
    :return: the historical configuration associated to the given time
    """
    try:
        for config_start_time_and_config in master_config[
            constants.CONFIG_HISTORICAL_CONFIGURATION
        ]:
            if config_start_time_and_config[0] <= current_time:
                return config_start_time_and_config[1]
        # no suitable config found: fallback to the oldest config
        return master_config[constants.CONFIG_HISTORICAL_CONFIGURATION][-1][1]
    except KeyError:
        raise KeyError(
            f"{constants.CONFIG_HISTORICAL_CONFIGURATION} not found in master_config."
        )


def get_oldest_historical_tentacle_config_time(master_config: dict) -> float:
    """
    :return: the oldest historical configuration timestamp
    """
    try:
        return min(
            historical_config[0]
            for historical_config in master_config.get(
                constants.CONFIG_HISTORICAL_CONFIGURATION, []
            )
        )
    except ValueError:
        raise ValueError("No historical configuration found")



# pylint: disable=R0913
#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc
import functools
import asyncio

import octobot_commons.logging as bot_logging
import octobot_commons.singleton as singleton


class Authenticator(singleton.Singleton):
    """
    Abstract class to be implemented when using authenticated requests
    """

    def __init__(self, use_as_singleton=True):
        self.logger: bot_logging.BotLogger = bot_logging.get_logger(
            self.__class__.__name__
        )
        self.initialized_event: asyncio.Event = None
        self.supports: None
        self.feed_callbacks = {}
        if use_as_singleton:
            # also register this instance for the base Authenticator class in singleton
            self.use_as_singleton_instance()

    # pylint: disable=W0212
    def use_as_singleton_instance(self):
        """
        Update the Authenticator Singleton to use self
        """
        singleton.Singleton._instances[Authenticator] = self
        singleton.Singleton._instances[self.__class__] = self

    @abc.abstractmethod
    async def login(self, username, password, password_token=None):
        """
        Used to trigger a login
        :param username: authentication username
        :param password: authentication password
        :param password_token: (optional) authentication password token
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def logout(self):
        """
        Used to clear a logged in session
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def register(self, username, password):
        """
        Used to create a user account
        :param username: authentication username
        :param password: authentication password
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_aiohttp_session(self):
        """
        Get the aiohttp authenticated session
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_logged_in(self):
        """
        :return: True when authenticated
        """
        raise NotImplementedError

    @abc.abstractmethod
    def must_be_authenticated_through_authenticator(self):
        """
        :return: True when this authenticator has to be validated
        """
        raise NotImplementedError

    @abc.abstractmethod
    def ensure_token_validity(self):
        """
        Called before @authenticated methods to ensure authentication
        :return:
        """
        raise NotImplementedError

    async def register_feed_callback(self, topic, callback):
        """
        Registers a feed callback
        """
        raise NotImplementedError

    async def send(self, message, channel_type, identifier=None):
        """
        Sends a message
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_initialized(self) -> bool:
        """
        Returns True when initialized
        :return:
        """
        raise NotImplementedError

    async def await_initialization(self, timeout):
        """
        Returns when initialized
        :return:
        """
        await asyncio.wait_for(self.initialized_event.wait(), timeout)

    async def update_trades(self, trades: list, exchange_name: str, reset: bool):
        """
        Updates authenticated account trades
        """
        raise NotImplementedError

    async def update_orders(self, orders_by_exchange: dict[str, list]):
        """
        Updates authenticated account orders
        """
        raise NotImplementedError

    async def update_positions(self, positions_by_exchange: dict[str, list]):
        """
        Updates authenticated account positions
        """
        raise NotImplementedError

    async def update_portfolio(
        self,
        current_value: dict,
        initial_value: dict,
        profitability: float,
        unit: str,
        content: dict,
        history: dict,
        price_by_asset: dict,
        reset: bool,
    ):
        """
        Updates authenticated account portfolio
        """
        raise NotImplementedError

    async def wait_for_private_data_fetch_if_processing(self):
        """
        Returns for private data to be fetched
        """

    def has_open_source_package(self) -> bool:
        """
        :return: True when open source package is available
        """
        raise NotImplementedError

    @staticmethod
    async def wait_and_check_has_open_source_package(raise_on_timeout=False) -> bool:
        """
        Returns for private data to be fetched and return True if package is available
        """
        authenticator = Authenticator.instance()
        try:
            await authenticator.wait_for_private_data_fetch_if_processing()
        except asyncio.TimeoutError:
            if raise_on_timeout:
                raise
        return authenticator.has_open_source_package()


class FailedAuthentication(Exception):
    """
    Raised upon authentication failure
    """


class UnavailableError(Exception):
    """
    Raised upon website availability issues failure
    """


class AuthenticationError(Exception):
    """
    Raised upon authentication technical error, not on login/password issues
    """


class AccountUpdateError(Exception):
    """
    Raised upon account update error
    """


class AuthenticationRequired(Exception):
    """
    Raised when an authentication is required
    """


def authenticated(func):
    """
    Annotation to required authentication for a method call
    :param func:
    :return:
    """

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        self.ensure_token_validity()
        return func(self, *args, **kwargs)

    return wrapped



# pylint: disable=W0718
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.logging as logging
import octobot_commons.configuration.user_inputs as user_inputs
import octobot_commons.databases.run_databases.run_databases_provider as run_databases_provider

try:
    import octobot_tentacles_manager.api
except ImportError:
    pass


def load_user_inputs_from_class(
    configured_class, tentacles_setup_config, to_fill_config
) -> dict:
    """
    Apply the given tentacles_setup_config configuration to the given to_fill_config using configured_class user inputs
    Requires octobot_tentacles_manager import, configured_class.CLASS_UI, configured_class.init_user_inputs
    and configured_class.get_name
    :return: the filled user input configuration
    """
    inputs = {}
    try:
        to_fill_config.update(
            octobot_tentacles_manager.api.get_tentacle_config(
                tentacles_setup_config, configured_class
            )
        )
    except NotImplementedError:
        # get_name not implemented, no tentacle config
        return inputs
    logger = logging.get_logger(configured_class.get_name())
    try:
        with configured_class.CLASS_UI.local_factory(
            configured_class, lambda: to_fill_config
        ):
            configured_class.init_user_inputs_from_class(inputs)
    except Exception as err:
        logger.exception(err, True, f"Error when initializing user inputs: {err}")
    if to_fill_config:
        logger.debug(f"Using config: {to_fill_config}")
    return inputs


async def load_and_save_user_inputs(tentacle_instance, bot_id: str) -> dict:
    """
    Requires an instance of the tentacle and the init_user_inputs method
    Initialize and save the user inputs of the tentacle in run data
    :return: the filled user input configuration
    """
    inputs = {}
    try:
        tentacle_instance.init_user_inputs(inputs)
        if run_databases_provider.RunDatabasesProvider.instance().is_storage_enabled(
            bot_id
        ):
            run_db = run_databases_provider.RunDatabasesProvider.instance().get_run_db(
                bot_id
            )
            await user_inputs.clear_user_inputs(run_db, tentacle_instance.get_name())
            for user_input in inputs.values():
                await user_inputs.save_user_input(user_input, run_db)
            await run_db.flush()
    except Exception as err:
        tentacle_instance.logger.exception(
            err, True, f"Error when initializing user inputs: {err}"
        )
    return inputs


def get_raw_config_and_user_inputs_from_class(
    configured_class, tentacles_setup_config
) -> (dict, list):
    """
    Requires octobot_tentacles_manager import and configured_class.load_user_inputs
    :return: the filled user input configuration of configured_class according to the given tentacles_setup_config
    """
    loaded_config = octobot_tentacles_manager.api.get_tentacle_config(
        tentacles_setup_config, configured_class
    )
    created_user_inputs = configured_class.load_user_inputs_from_class(
        tentacles_setup_config, loaded_config
    )
    return loaded_config, list(
        user_input.to_dict() for user_input in created_user_inputs.values()
    )


async def get_raw_config_and_user_inputs(
    configured_class, config, tentacles_setup_config, bot_id
) -> (dict, list):
    """
    Requires octobot_tentacles_manager import and configured_class.create_local_instance
    Uses run data to load user input values when available
    :return: the tentacle configuration and its list of user inputs
    """
    loaded_config = octobot_tentacles_manager.api.get_tentacle_config(
        tentacles_setup_config, configured_class
    )
    if saved_user_inputs := await user_inputs.get_user_inputs(
        run_databases_provider.RunDatabasesProvider.instance().get_run_db(bot_id),
        configured_class.get_name(),
    ):
        # user inputs have been saved in run database, use those as they might contain additional
        # (nested) user inputs
        return loaded_config, saved_user_inputs
    # use user inputs from init_user_inputs
    tentacle_instance = configured_class.create_local_instance(
        config, tentacles_setup_config, loaded_config
    )
    created_user_inputs = {}
    tentacle_instance.init_user_inputs(created_user_inputs)
    return loaded_config, list(
        user_input.to_dict() for user_input in created_user_inputs.values()
    )
